// Electric Parking System (EPS) Override
// this code is used to override the EPS that was bought from the scrap yard
// 
// rpascoal@ua.pt
// Copyright LARLABS
//
// read digital input to check for actuator request
// read the force sensor
// choose the motor direction
// activate the Mosfet with PWM until the desired force is met or exceeded
//
// !!!Force scale is not linear!!!
// Null force results in a voltage rms of 1.1 V
// EPS active results in a voltage rms greater than 1.3 V (only tested roughly up to 26kgf)
// give 100mV to max and min => must work with rms
// !!!Force scale is not linear!!!

#ifndef MAIN_C
#define MAIN_C

#include <p18f258.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "main.h"
//#include <pwm.h>						// MPlab PWM library functions
//#include <adc.h>						// MPlab ADC library functions


// This can also be done in MPLAB configuration bits
// Configure the device
// check page 265 of the device datasheet "SPECIAL FEATURES OF THE CPU"
// most can be left unprogrammed
//#pragma config OSC = HS // HS Oscillator
//#pragma config OSCS = OFF // Oscillator switching
//#pragma config PWRT = ON // Power on reset
//#pragma config BOR = ON //brown-out reset
//#pragma config BORV = 27 // brown out at 2.7V
//#pragma config WDT = OFF // watch dog off
//#pragma config STVR = OFF // stack overflow reset
//#pragma config LVP = OFF // low voltage programming
//#pragma config DEBUG = OFF // background debug
//
//#pragma config CP0 = OFF               //Code Protection Block 0:Disabled
//#pragma config CP1 = OFF               //Code Protection Block 1:Disabled
//#pragma config CP2 = OFF               //Code Protection Block 2:Disabled
//#pragma config CP3 = OFF               //Code Protection Block 3:Disabled
//#pragma config CPB = OFF               //Boot Block Code Protection:Disabled
//#pragma config CPD = OFF               //Data EEPROM Code Protection:Disabled
//#pragma config WRT0 = OFF              //Write Protection Block 0:Disabled
//#pragma config WRT1 = OFF              //Write Protection Block 1:Disabled
//#pragma config WRT2 = OFF              //Write Protection Block 2:Disabled
//#pragma config WRT3 = OFF              //Write Protection Block 3:Disabled
//#pragma config WRTB = OFF              //Boot Block Write Protection:Disabled   
//#pragma config WRTC = OFF              //Configuration Register Write Protection:Disabled
//#pragma config WRTD = OFF              //Data EEPROM Write Protection:Disabled
//#pragma config EBTR0 = OFF             //Table Read Protection Block 0:Disabled
//#pragma config EBTR1 = OFF             //Table Read Protection Block 1:Disabled
//#pragma config EBTR2 = OFF             //Table Read Protection Block 2:Disabled
//#pragma config EBTR3 = OFF             //Table Read Protection Block 3:Disabled
//#pragma config EBTRB = OFF             //Boot Block Table Read Protection:Disabled


void high_isr(void);

#pragma code high_vector=0x08
	void interrupt_at_high_vector (void)
	{
	 _asm GOTO high_isr _endasm
	}
#pragma code


#pragma interrupt high_isr
void high_isr(void)
{

// analogue conversion finished
//if(ADCON)
//{
//}


// if a rising edge interrupt is detected in the digital input
// check moving average value of the force sensor
if(INTCONbits.INT0IF)
{
	IsUndefined = 1; // to make sure that positions other than locked or unlocked are taken care of by locking
	if (Force > ForceToLock)
	{
		IsLocked = 1;
		IsUnlocked = 0;
		IsUndefined = 0;
	}
	if (Force < ForceToUnlock)
	{
		IsLocked = 0;
		IsUnlocked = 1;
		IsUndefined = 0;
	}

	// already locking or unlocking
	// this can also be done by disabelling the interrupt on the INT0
	if(IsUnlocking || IsLocking)
	{
	  // do not answer the interrupt request
	 RequestedPositionReached = 0;
	  
	}
	else
	{
// if eps is in the locked position then switch relay to unlock and activate the mosfet
// when unlocked position is reached deactivate the mosfet and the relay and allow locking
	if (IsLocked)
	{	
	  LockActionRequired = 0;
	  UnlockActionRequired = 1;
	  //IsUnlocking = 1;
	}
// if eps is in the unlocked position then switch relay to lock and activate the mosfet
// when locked position is reached deactivate the mosfet and the relay and allow unlocking
// as safety measure: if for some reason the eps is halfway, and the motor is off, then go to locked position
	if (IsUnlocked || IsUndefined)
	{
		
		UnlockActionRequired = 0;
		LockActionRequired = 1;
		//IsLocking = 1;
	}


	}; //if else
	
	//page 79 of datasheet says INT0 is cleared by reading port B
	INTCONbits.INT0IF=0; // clear INT0 interrupt
	temp = PORTB;
	

  }	
} //high_isr


void Setup(void)
{
	UnlockActionRequired = 0;
	LockActionRequired = 0;
	IsLocking = 0;
	IsUnlocking = 0;

	MA_ = 0;
	//dutycyle = 0;

	// values from the force sensor

	ForceToLock = 270; // corresponding to 16 kg and 1.32 Volt CHECK THIS
	ForceToUnlock = 120; //CHECK THIS, just a testing value
	ForceThreshold = 5; // CHECK

	// Inputs and Outputs
	TRISA = 0b1111111; // all input
	TRISB = 0b11111111; // all input
	TRISC = 0b11100000; //
	PORTCbits.RC4 = 1; //alive
	RelayUnlock = 0; // guarantee off
	RelayLock = 0; // guarantee off


	//First two are analogue input
	ADCON1bits.PCFG0=0;
	ADCON1bits.PCFG1=0;
	ADCON1bits.PCFG2=1;
	ADCON1bits.PCFG3=0;
	ADCON1bits.ADFM =1;				// Right justified A/D result
	// select internal adc clock
	ADCON1bits.ADCS2 = 0;
	ADCON0bits.ADCS0 = 1;
	ADCON0bits.ADCS1 = 1;

	ADCON0bits.CHS0 = 0; //channel AN0
	ADCON0bits.CHS1 = 0; 
	ADCON0bits.CHS2 = 0; 

	ADCON0bits.ADON = 1; //turn on the converter

	//
	CCP1CONbits.CCP1M3 = 1; // PWM mode
	CCP1CONbits.CCP1M2 = 1; //   <<
	PR2 = 49;				// Sets PWM Period, Fosc = 4MHZ
	T2CON = 0;				// TMR2 Off with prescale and postscale = 1
	CCPR1L = 0;				// Sets Duty Cycle to zero
	CCP1CONbits.DC1B0 = 0;
	CCP1CONbits.DC1B1 = 0;
	T2CONbits.TMR2ON = 1;	// Start Timer2

	//set INT0 for interrupt on rising edge
	INTCON2bits.INTEDG0 = 1;
	// INT0 is always high priority

	//Activate interrupts
	INTCONbits.GIE=1;		//global interrupt enable
	INTCONbits.INT0IE=1;	//enable interrupt INT0
}

void MosfetPWM(unsigned int dutycyle)
{
 	if(dutycyle == 0)
	{
//		CCP1CONbits.CCP1M3=0x0;	// turn off the PWM generation
//		CCP1CONbits.CCP1M2=0x0;
//		CCP1CONbits.CCP1M0=0x0;
//		CCP1CONbits.CCP1M1=0x0;
		CCP1CONbits.DC1B0 = 0;
		CCP1CONbits.DC1B1 = 0;
		CCPR1L = 0;
 	}
	else
	{  	
		CCP1CONbits.DC1B1 = 0b00000001 & (dutycyle>>1); //(dutycyle%4)>>1
		CCP1CONbits.DC1B0 = 0b00000001 & dutycyle; //((dutycyle%4)%2)
		CCPR1L = dutycyle>>2; //8MSbs of 10
	}	
}

unsigned int runavg(unsigned int x, unsigned int N)
{	// low memory usage moving average
	MA_ = MA_ + x - MA_/N;
	return MA_/N;
}

void delay(unsigned int delayval)
{
	while(delayval)
	{
	 delayval--;
	}
}

unsigned int ReadForce(void)
{	
	unsigned int readval;
	readval = 0;
	ADCON0bits.CHS0 = 0; //channel AN0
	ADCON0bits.CHS1 = 0; 
	ADCON0bits.CHS2 = 0; 
	delayval = 100; // for 4MHz OSC, 1MHz instruction 1us
	delay(delayval); // capacitor charged
	ADCON0bits.GO_DONE = 1; // start conversion
	while(ADCON0bits.GO_DONE)
	{
	  delayval = 0; // waist time; still converting
	}
	readval = ADRESH * 256;		// the A/D result
	readval = readval + ADRESL;
	return readval;
}

void main(void)
{
  // wait some time to start the loop code so that the influence of voltage spiking is eliminated
  // this can also be met with an external power on reset
   
  delay(10);
  Setup();
  entered_once = 0;
  while(1)
  {

   //continuously read the force sensor and perform a moving average that is available if necessary
	Force_i = ReadForce();
 	// this is not a 
	Force = runavg(Force_i, 64); // 64 samples to average
	

	if(LockActionRequired)
	{ 
	  IsLocking = 1;

	  if(entered_once == 0)
	  {
		RequestedPositionReached = 0;
	  	RelayUnlock = 0; // guarantee off
	  	RelayLock = 1; // switch to unlock
	    delay(100); // time for the relay to close
		entered_once = 1;
		MosfetPWM(1023); // full on for now, take outside the if for control over the PWM
	  } 
	  if (Force > (ForceToLock + ForceThreshold))
		{
		MosfetPWM(0);
		RelayLock = 0; // switch off, no need for delay because MOSFET is fast
		LockActionRequired = 0;
		RequestedPositionReached = 1;
		entered_once = 0;
		IsLocking = 0; //finished locking
		}
	}

	if(UnlockActionRequired)
	{
	  IsUnlocking = 1;
	 if(entered_once == 0)
	 {	
	  RequestedPositionReached = 0;
	  RelayLock = 0; // guarantee off
	  RelayUnlock = 1; // switch to unlock
	  delay(100); // time for the relay to close
	  MosfetPWM(1023); // full on for now
	 }

	  if (Force < (ForceToUnlock - ForceThreshold))
		{
		 MosfetPWM(0); // turn gate off
		 RelayUnlock = 0; // switch off
		 UnlockActionRequired = 0;
		 RequestedPositionReached = 1;
		 IsUnlocking = 0; //finished unlocking
		}
	}

  }
}

#endif
